require 'rubygems'
require 'yaml'
require 'optparse'
require 'amqp'
require 'json'


# The program manages job commands. Master nodes (workflow handler) sends
# messages to jobhandler instances. Each instance manages one and
# only one message at a time. Jobhnadlers are also in charge of storage to S3.
# Author: Olivier Sallou <olivier.sallou@irisa.fr>
# Copyright: 2012, IRISA


if ENV["MYSQL_URL"]==nil
  puts "MYSQL_URL environment variable is not set exiting..."
  exit 1
end
if ENV["AMQP_URL"]==nil
  puts "AMQP_URL environment variable is not set, using defaults."
end

$:.push File.expand_path("../lib")

require 'manband/job.rb'
require 'manband/store.rb'

handler = Time.now.to_s

log = Logger.new(STDOUT)
log.level = Logger::INFO

options = {}

optparse = OptionParser.new do|opts|
  
  # This displays the help screen, all programs are
  # assumed to have this option.
  opts.on( '-h', '--help', 'Display this screen' ) do
    puts opts
    exit
  end

  opts.on( '-i', '--id ID', 'Handler id' ) do |id|
    handler = id.to_s
  end

  options[:queue] = ""
  opts.on( '-q', '--queue QUEUE', 'Queue registration filter' ) do |queue|
    options[:queue] = "."+queue
  end

  options[:store] = nil
  opts.on( '-s', '--store S3HOST', 'S3 host information host:port:path, e.g. localhost:8773:/services/Walrus' ) do |store|
    options[:store] = store
  end

  options[:conf] = nil
  opts.on( '-c', '--conf CONFIG', 'Use config file' ) do |config|
    options[:conf] = config
  end

end

optparse.parse!

userconf = File.expand_path("~")+"/.manband"
if options[:conf]==nil && File.exists?(userconf)
  options[:conf]=userconf
end

if options[:conf]!=nil
  if !File.exists?(options[:conf])
    puts "Config file does not exists"
    exit 1 
  end
  conf = YAML.load_file(options[:conf])
  puts "Using configuration file "+options[:conf]
  FlowConfig.sets3(conf["s3"]["host"],conf["s3"]["port"],conf["s3"]["path"])
  FlowConfig.setworkdir(conf["workdir"])
  if conf["sudo"] != nil
    log.info "Sudo mode:"+conf["sudo"].to_s
    FlowConfig.setsudo(conf["sudo"])
  end
end

if options[:store]!=nil
  elts = options[:store].split(':')
  FlowConfig.sets3(elts[0],elts[1],elts[2])
end

log.info "Start handler "+handler

urlconfig = FlowConfig.config()


EventMachine.run do
  connection = AMQP.connect(FlowConfig.config())

  show_stopper = Proc.new do
    $stdout.puts "Stopping..."
    connection.close {
      EventMachine.stop { exit }
    }
  end

  Signal.trap 'INT', show_stopper
  Signal.trap 'TERM', show_stopper

  channel_exception_handler = Proc.new { |ch, channel_close| EventMachine.stop; raise "channel error: #{channel_close.reply_text}" }

  channel    = AMQP::Channel.new(connection)
  #channel.on_error(&channel_exception_handler)
  channel.prefetch(1)

  exchange = channel.direct()
  queue = channel.queue('manband.node'+options[:queue], :auto_delete => false, :durable => true)
  queue.bind(exchange).subscribe(:ack => true) do |metadata, message|
    log.debug "Received msg: "+message.to_s
    doack = true
    args = JSON.parse(message)
    msg = JSON.parse(args["msg"])
    if args["operation"] == OP_DESTROY
      myworkflow = WorkFlow.get(msg["id"])
      if myworkflow != nil
        myworkflow.delete()
      end
    elsif args["operation"] == OP_CLEAN
      myworkflow = WorkFlow.get(msg["id"])
      if myworkflow == nil
        log.error "Error: workflow does not exists "+msg["id"]
      else
        myworkflow.clean
      end
    else
      myjob = Job.get(msg["id"])
      if myjob == nil
        log.error "Error: job does not exists "+msg["id"]
      else 
        if args["operation"] == OP_RUN
          myjob.setAck(channel,metadata.delivery_tag)
          doack = false
          if msg["instance"]!=nil
            myjob.run(handler, msg["instance"].to_i)
          else
            myjob.run(handler)
          end
        elsif args["operation"] == OP_SKIP
          myjob.skip(handler)
        elsif args["operation"] == OP_STORE
          storage = Storeband.new
          uid = WorkFlow.get(myjob.wid).uid
          storage.store(myjob,uid,msg["bucket"])
        end
      end
    end


    if doack
      channel.acknowledge(metadata.delivery_tag, false)
    end
  end

end

