require 'rubygems'
require 'yaml'
require 'optparse'
require 'amqp'
require 'json'


# The program s the main orchestrator of the workflow.
# It starts with a workflow execution request and sends messages to job
# handlers according to the input workflow.
# It is possible to have multipe workflow handler, sharing messages for 
# the same workflow, or to handler a high number of workflow requests.
# Author: Olivier Sallou <olivier.sallou@irisa.fr>
# Copyright: 2012, IRISA


if ENV["MYSQL_URL"]==nil
  puts "MYSQL_URL environment variable is not set exiting..."
  exit 1
end
if ENV["AMQP_URL"]==nil
  puts "AMQP_URL environment variable is not set, using defaults."
end

$:.push File.expand_path("../lib")

require 'manband/flowconfig.rb'
require 'manband/workflow.rb'
require 'manband/job.rb'
require 'manband/bandmanager.rb'

log = Logger.new(STDOUT)
log.level = Logger::INFO


@@options = {}

def savemessage(id,operation,msg)
   if @@options[:debug] == false
     return
   end
   bmsg = BandMessage.new(:wid => id, :message => '{ "operation" => '+operation+', "msg" => '+msg+' }' )
   bmsg.save
end

optparse = OptionParser.new do|opts|
  
  # This displays the help screen, all programs are
  # assumed to have this option.
  opts.on( '-h', '--help', 'Display this screen' ) do
    puts opts
    exit
  end

  @@options[:conf] = nil
  opts.on( '-c', '--conf CONFIG', 'Use config file' ) do |config|
    @@options[:conf] = config
  end


  @@options[:debug]=false
  opts.on( '-d', '--debug', 'Debug mode, store all messages in database') do
    @@options[:debug]=true
    Job.debug(true)
  end
end

optparse.parse!

userconf = File.expand_path("~")+"/.manband"
if @@options[:conf]==nil && File.exists?(userconf)
  @@options[:conf]=userconf
end

if @@options[:conf]!=nil
  if !File.exists?(@@options[:conf])
    puts "Config file does not exists"
    exit 1
  end
  conf = YAML.load_file(@@options[:conf])
  puts "Using configuration file "+@@options[:conf]
  FlowConfig.sets3(conf["s3"]["host"],conf["s3"]["port"],conf["s3"]["path"])
  FlowConfig.setworkdir(conf["workdir"])
end


EventMachine.run do
  connection = AMQP.connect(FlowConfig.config())

  show_stopper = Proc.new do
    $stdout.puts "Stopping..."
    connection.close {
      EventMachine.stop { exit }
    }
  end

  Signal.trap 'INT', show_stopper
  Signal.trap 'TERM', show_stopper

  channel_exception_handler = Proc.new { |ch, channel_close| EventMachine.stop; raise "channel error: #{channel_close.reply_text}" }

  channel    = AMQP::Channel.new(connection)
  #channel.on_error(&channel_exception_handler)
  channel.prefetch(1)

  exchange = channel.direct()
  queue = channel.queue('manband.master', :auto_delete => false, :durable => true)
  queue.bind(exchange).subscribe(:ack => true) do |metadata, message|

    begin
    log.debug "Received msg: "+message.to_s
    args = JSON.parse(message)
    msg = JSON.parse(args["msg"])
    rescue Exception => e
      log.error("Could not parse message "+args["msg"])
      channel.acknowledge(metadata.delivery_tag, false)
    end
    if args["operation"] == OP_NEW
      BandManager.launch(msg["wfile"],msg["var"],msg["uid"],msg["bucket"])
    end
    if args["operation"] == OP_START
      savemessage(msg["id"],args["operation"],args["msg"])
      workflow = WorkFlow.get(msg["id"])
      if workflow == nil
        log.error "Error: workflow does not exists "+msg["id"] 
      else
        workflow.update(:status => STATUS_RUNNING)
        myjob = Job.get(msg["root"])
        myjob.runnext
      end
    end
    if args["operation"] == OP_SKIP
      savemessage(msg["id"],args["operation"],args["msg"])
      workflow = WorkFlow.get(msg["workflow"])
      if workflow == nil
        log.error "Error: workflow does not exists "+msg["workflow"]
      else
        workflow.update(:status => STATUS_RUNNING)
        myjob = Job.get(msg["id"])
        myjob.runnext(true)
      end
    end
    if args["operation"] == OP_ERROR
      savemessage(msg["workflow"],args["operation"],args["msg"])
      myjob = Job.get(msg["id"])
      if myjob == nil
        log.error "Error: job does not exists "+msg["id"]
      else
        log.debug "Job "+msg["id"]+" in error"
        instance = 0
        if msg["instance"]!=nil
          instance = msg["instance"]
        end
        myjob.error!(instance)
      end
    end
    if args["operation"] == OP_FINISH
      savemessage(msg["workflow"],args["operation"],args["msg"])
      myjob = Job.get(msg["id"])
      doSkip = false
      if myjob == nil
        log.error "Error: job does not exists "+msg["id"]
      else
        if myjob.status == STATUS_SKIP
          doSkip = true
        end
        if msg["handler"]!=nil && myjob.handler != msg["handler"] && myjob.type!=INTERACTIVE_ACTOR
          log.debug "Finish received from a different handler "+myjob.handler+" vs "+msg["handler"]+", skipping message"
        else
          myjob.finish
          if myjob.isover?
            myjob.runnext(doSkip)
          end
        end
      end
    end
    if args["operation"] == OP_WSUSPEND
      savemessage(msg["id"],args["operation"],args["msg"])
      workflow = WorkFlow.get(msg["id"])
      if workflow == nil
        log.error "Error: workflow does not exists "+msg["id"]
      else
        workflow.update(:status => STATUS_SUSPEND)
      end
    end
    if args["operation"] == OP_WRESUME
      savemessage(msg["id"],args["operation"],args["msg"])
      workflow = WorkFlow.get(msg["id"])
      if workflow == nil
        log.error "Error: workflow does not exists "+msg["id"]
      else
        log.info "Resuming workflow "+msg["id"]
        workflow.update(:status => STATUS_RUNNING)
        suspendedjobs = Job.all(:status => STATUS_SUSPEND, :wid => workflow.id)
        suspendedjobs.each do |job|
          job.resume
        end
        errorjobs = Job.all(:status => STATUS_ERROR, :wid => workflow.id)
        errorjobs.each do  |job|
          job.resume
        end
      end
    end
    if args["operation"] == OP_JSUSPEND
      savemessage(msg["workflow"],args["operation"],args["msg"])
      curjob = Job.get(msg["id"])
      if curjob == nil
        log.error "Error: Job does not exists "+msg["id"]
      else
        if curjob.status != STATUS_SKIP
          curjob.update(:status => STATUS_SUSPEND)
        end
      end
    end
    if args["operation"] == OP_JRESUME
      savemessage(msg["workflow"],args["operation"],args["msg"])
      curjob = Job.get(msg["id"])
      if curjob == nil
        log.error "Error: Job does not exists "+msg["id"]
      else
        log.info "Resuming job "+msg["id"]
        curjob.resume
      end
    end


  channel.acknowledge(metadata.delivery_tag, false)
  end

end

